
import java.io.*;
import java.util.*;

/**
 * 
 Program to compute the W-set for a given FSM.
 Based on Chow's algorithm published in 1980
 This program is based on some code written by Aditya for the UIO algorithm.
 
 Note that the algorithm does NOT work for machines with self loops in states.
 An FSM without any self loops represents  the core behavior of an FSM.
 
 Originally coded by: Aditya Mathur (for the UIO algorothm. Some of those files are used here).
 W-method and test sets coded by a bright Purdue undergraduate students.
 
 Comments:Variables in the initial version are the same as given by the authors of the algorithm.
 Java naming norms may be  violated.

 */


public class wpAlgorithm{
  
  private static String debugMode1="";  // No debugging.
  private static String debugMode2="";
  public static final int maxStates=20;  // Max no of states in the FSM (=n).
  public static final int maxSequences=100;  // Max no of sequences of length (2xnxn).
  public static final int maxEdges=100;  // Max no of edges in the FSM.
  public static final int maxOutgoingEdges=20;  // Max no of outgoing edges/state.
  public static final int maxInputs=50;  // Max ize of the input alphabet.
  public static final int maxOutputs=50;  // Max size of the output  alphabet.
  public static final int maxTransitions=250;  // Max number of transitions.
  public static int maxLabels=50; // Max number of distinct edge labels.
  
  // An FSM state contains an ID, description, and a set of outgoing edges.
  public static State [] FSM =new State[maxStates]; // Area where the FSM will be stored.
  public static int numberOfStates; // Number of distinct states in the FSM.
  public static int startState; // Starting state of the FSM.
  
  
  Set inputAlphabet=new HashSet();// Contains input alphabet symbols.
  Set  outputAlphabet=new HashSet(); // Output alphabet symbols.
  public static int numberOfTransitions=0; // Total transitions in the FSM.
  public static String fsmFilename; // File containing the FSM description.
  

  public static int numberOfTransitions2 = 0;
  public static String fsmFilename2;
  
  private static String [] Opattern=new String[maxEdges];; // Holds all patterns being tried out.
  private static int [] Oend=new int[maxEdges]; // Holds end tail state corresponding to each pattern in Opattern.
  private static Set [] Oled=new HashSet[maxEdges]; // Holds sets of edges corresponding to a labels.
  
  
  private static String[] outputArray = new String[maxOutputs]; //Output alphabet
  private static String[] inputArray = new String[maxInputs]; // Input alphabet
  private static int countOutputs = 0; //Size of the output alphabet
  private static int countInputs = 0; // Size of the input alphabet
  
  private static Scanner fileSource; // Name of a Scanner object.
  
  /*
   * 
   The next few variables are computed by the UIO algorithm.
   
   */
  private static LabelAndEdgeSet [] distinctEdgeLabels=new LabelAndEdgeSet[maxLabels];  // Set of distinct edge labels.
  private static int numberOfDistinctEdgeLabels=0;
  
  /**
   * 
   Method to read the FSM description from a source file.
   startState is assumed to be 1.
   */
  
  public static void getFSM(){
    String aTransition; // One transition from input.
    int currentState=-1; // ID of state being input (-1 is an invalid ID)
    numberOfStates=0; // No states input so far.
    startState=1;
    try{
      BufferedReader fsmFile= new BufferedReader (new FileReader(fsmFilename));
      aTransition=fsmFile.readLine();
      boolean done=false; // Indicates loop to be terminated or not.
      if (aTransition==null) 
        done=true;
      
      while (!done){
        numberOfTransitions++;  // One more transition input.
        //   System.out.println("Next transition: "+numberOfTransitions+" "+aTransition);
        splitTransition(aTransition);  // Separate from, to, and edge label of the transition and add to FSM.
        aTransition=fsmFile.readLine(); // Get the next transition.
        if (aTransition==null || (numberOfTransitions==maxTransitions)) 
          done=true;
      }
    }catch (FileNotFoundException e){
      System.out.println("File: "+fsmFilename+" not found");
      System.exit(0);
    }catch (IOException e){
      Utilities.printException("wPAlgoriithm", "getFSM","IO exception raised while reading: "+
                               fsmFilename+" not found");
      Utilities.printException("wPAlgoriithm", "getFSM", "Transitions input: "+numberOfTransitions);
    }
  } //End getFSM()
  
  
  
  static void splitTransition(String transition){
    
    /*
     * Split a line of input from the FSM file.
     * Example:
     * Input line: 3 4 a/0  Note the three tokens 3, 4, and a/0.
     * Indicates: Source state: 3
     * Destination state:  4 
     * Take the transition on input: a
     * Generate output: 0.
     */
  
    String [] token =new String[3];
    String [] edgeToken =new String[2];
    String inputLabel, outputLabel, toState, edgeLabel; // Temporaries.
    int ID; // ID of the from state in a transition.
    int toID; //ID of the tdestination satte in a transition.
    
    StringTokenizer transitionTokens=new StringTokenizer(transition);
    int tokensFound=0;
    
    // Split transition into its three components: source, destination and edge label.
    while(transitionTokens.hasMoreTokens()){
      token[tokensFound]=transitionTokens.nextToken();
      Utilities.debugFSM("Next token: "+token[tokensFound]);
      tokensFound++;
    }// End of while
    
    ID=-1; // To avoid compiler error.
    try{
      ID=Integer.parseInt(token[0]);  // Convert from state ID to integer.
      if (newState(ID)){ // Check if this state did not appear before.
      numberOfStates++;
      FSM[ID]=new State(ID);  // Create a new state entry in the FSM.
    }
    }catch(NumberFormatException e){
      Utilities.printException("wpAlgorithm", "splitTransition", "From state ID must be a number, found: "+token[0]);
    }
    
    inputLabel=""; // Initilized to avoid copmpiler error.
    outputLabel=""; // Initilized to avoid copmpiler error.
    toID=-1;
    try{
      toID=Integer.parseInt(token[1]);  // Destination state.
      edgeLabel=token[2]; // Edge label
      edgeToken=splitEdgeLabel(edgeLabel);  // Split edge label to input and output labels.
      inputLabel=edgeToken[0]; // Set input and 
      outputLabel=edgeToken[1]; // output labels.
    }catch(NumberFormatException e){
        Utilities.printException("wpAlgorithm", "splitTransition", "To state ID must be a number, found: "+token[1]);
      }
    
    // Check whether or not the output label has already appeared.
    int traverser = 0;
    boolean found = false;
    while(traverser < outputArray.length){
      if(outputArray[traverser].equals(outputLabel)){ // Check if symbol already in the output array.
        found = true;
        break;
      }
      traverser++; 
    }
    // Add output symkbol to the array if it does not already exist.
    
    if(found == false && !outputLabel.equals("")){
      Utilities.debugFSM("Adding " + outputLabel + " to output array position " + countOutputs + ".");
      outputArray[countOutputs] = outputLabel;
      countOutputs++;
    }
    
  // Check whether or not the input label has already appeared.
    found = false;
    traverser = 0;
    int traverser2 = 0;
    boolean found2 = false;
    while(traverser2 < inputArray.length){
      if(inputArray[traverser2].equals(inputLabel)){ //already in the input array.
        found2 = true;
        break;
      }
      traverser2++;
    }
    
    // Add input symbol to the input array if it does not already exist.
    if(found2 == false && !inputLabel.equals("")){
      Utilities.debugFSM("Adding " + inputLabel + " to input array position " + countInputs + ".");
      inputArray[countInputs] = inputLabel;
      countInputs++;
    }
    
    found2 = false;
    traverser2 = 0;
    
    
    try{
      FSM[ID].addEdge(new Edge(ID, toID, inputLabel, outputLabel));
    }catch (InvalidEdgeException e){
      Utilities.debugFSM("Trying to add an invalid edge. State ID does not match head ID");
      Utilities.debugFSM("State ID:" + ID+ " Edge head ID" + ID); // Impossible here.
      System.exit(0);
    }
  } // End splitTransition()
  
  /**
   Method to split an edge label of the form x/y to x and y.
   */
  
  private static String []  splitEdgeLabel(String edgeToken){
    String []token=new String [2];
    Utilities.debugFSM("EDGE TOKEN:" + edgeToken); 
    StringTokenizer ioTokens=new StringTokenizer(edgeToken, "/");
    token[0]=ioTokens.nextToken();
    token[1]=ioTokens.nextToken();
    return token;
  }//End splitEdgeLabel()
  
  
  /**
   * 
   Find if a given state ID corresponds to a new state in the FSM.
   */
  private static boolean newState(int stateID){
    if (FSM[stateID]!=null){
      return(false);
    }
    else{
      return(true);
    }
  }//End newState()
  
  
  
  /**
   Method to print all states and transitions of the input FSM.
   */
  
  public static void printFSM(String [] inputAlphabet){
    // Temporary variables.
    
    Set edges=new HashSet();
    int stateID;
    
    System.out.println("States: "+numberOfStates);
    System.out.println("Edges "+numberOfTransitions);
    System.out.println("Input alphabet:");
    for (int i=0; i<inputAlphabet.length; i++)
      System.out.println(inputAlphabet[i]);
    
    System.out.println("\nOutput alphabet:");
    int count = 0;
    Arrays.sort(outputArray);
    while(count<maxOutputs){
      if(!outputArray[count].equals(""))
        System.out.println(outputArray[count]); 
      count++;
    }
    
    
    System.out.println("\nFrom \t Input/Output \t To");
    
    for (int i=0; i<maxStates; i++){
      if(FSM[i]!=null){
        stateID=FSM[i].getID();
        edges=FSM[i].getEdgeSet();
        Iterator E=edges.iterator();
        while(E.hasNext()){
          Edge e=(Edge)E.next();
          System.out.println(stateID+ "\t "+ e.input()+ "/" + e.output()+ "\t\t "+ e.tail());
        }// End of while over edges from a node
      }// End of if to check ith state.
    }// End of for to iterate over all states.
    
    
    
    count = 0;
   // Arrays.sort(inputArray);
    sortInputs();
    while(!inputArray[count].equals("")){
      Utilities.debugFSM("Possible input "+count +": " +inputArray[count]); 
      count++;
    }
  }// End printFSM()
  
  
  
  private static String getFilename(){
    System.out.print("Enter filename: "); // Prompt for file name containing FSM.
    String name=fileSource.nextLine();
    return (name);
  }//End getFileName()
  
  
  /**
   * 
   Obtain distinct edge labels for the FSM. For each unique label compute the set of edges.
   */
  
  private static void getDistinctEdgeLabels(){
    Set edges=new HashSet();
    HashSet nextStateSet=new HashSet();
    LabelAndEdgeSet label;  // Label in one element of distinctLabelAndEdgeSet;
    
    numberOfDistinctEdgeLabels=0;
    if(!debugMode1.equals(""))System.out.println("Finding distinct edge labels for state:");
    for (int i=0; i<maxStates; i++){
      if(FSM[i]!=null){
        int numberOfEdges=FSM[i].edgeCount();
        edges=FSM[i].getEdgeSet(); // All edges out of state i.
        Iterator E=edges.iterator();
        if(!debugMode1.equals(""))System.out.println("State:"+i);
        while (E.hasNext()){
          Edge e=(Edge)E.next();
          String nextEdgeLabel=e.getLabel();
          if(!debugMode1.equals(""))System.out.println("Next edge label considered:"+nextEdgeLabel);
          boolean found=false;
          int index=labelIn(nextEdgeLabel, numberOfDistinctEdgeLabels);
          if(index==-1){
       // Not found. New label.
            distinctEdgeLabels[numberOfDistinctEdgeLabels]=new LabelAndEdgeSet();
            distinctEdgeLabels[numberOfDistinctEdgeLabels].addLabel(nextEdgeLabel);
            
            distinctEdgeLabels[numberOfDistinctEdgeLabels].addEdge(e);
            numberOfDistinctEdgeLabels++;
          }else{
            distinctEdgeLabels[index].addEdge(e);
          }
        }
      }
    }
  }// End of getDistinctEdgeLabels()
  
  /**
   Check if a given label has already been discovered. Return ndex if found else -1.
   */
  
  private  static int labelIn(String l, int count){
    
    if(!debugMode1.equals(""))System.out.println("Checking if label already found ");
    for (int i=0; i<count; i++){
      LabelAndEdgeSet s=distinctEdgeLabels[i];
      String label=s.getLabel();
      if(label.equals(l)) 
        return(i);
    } 
    return(-1);
  }// End of labelIn()
  
  
  
  
  private static void printEdges(Set s){
    Iterator E=s.iterator();
    if(!E.hasNext())
    {
      System.out.println("Set to be printed is empty.");
    }
    else
    {
      while(E.hasNext())
      {
        Edge e=(Edge)E.next();
        System.out.println(e.head()+" "+ e.tail()+" "+e.getLabel());
      }
    }
  }//End of printEdges()
  
  /**
   * 
   Copy all elements of set "from" to set "to".
   
   */
  
  private static void makeCopy( Set from, Set to){
 
    to.clear();  // Empty the destination set.
    Iterator E=from.iterator();
    while (E.hasNext()){
      Edge e=(Edge)E.next();
      to.add(e);
    }
  }// End of makeCopy()
  
  
  private static int findLabelIndex(String label){
    Set edges=new HashSet();
    LabelAndEdgeSet lab=new LabelAndEdgeSet();
    boolean labelExists=false; // label must exist as a distinct label in the distinctEdgeLabels.
    for (int i=0; i<numberOfDistinctEdgeLabels; i++){
      lab=distinctEdgeLabels[i]; // Label set associated with the ith distinct label.
      edges=lab.getLabelEdges(); // This is the set of all edges associated with lab
      if(lab.getLabel().equals(label))return(i);
    }
    return(-1); // Label not found. This must not happen.
  }// End of findLabelIndex().
  
  /**
   Utility routine to display a list of edges.
   
   */
  
  public static void displayEdges(Set edges){
    Iterator E=edges.iterator();
    while(E.hasNext()){
      Edge e=(Edge)E.next();
      System.out.println( " Edge: "+e.head()+"\t"+e.tail()+"\t"+e.getLabel());
    }
    
  }
  
  
  public static void runFSM(int stateID, String input, String separator){
    
    // input is a sequence of input symbols separated by string in separator.
    // StateId is the ID of the state to which input is to be applied.
    String outputPattern="";  // Output sequence generated by the FSM. Illegal token generates empty output.
    String token;  // An input symbol from the input sequence.
    StringTokenizer inputTokens=new StringTokenizer(input, separator);
    int currentState=stateID;  // Rest the FSM to state StateID.
    Utilities.debugFSMExecution("FSM execution begins");
    if(FSM[stateID]==null){
      Utilities.printException("wpAlgoriothm", "runFSM", "Invalid start state. Execution aborted.");
      return;
    }
    while(inputTokens.hasMoreTokens()){
      token=inputTokens.nextToken(); //Get next token from input.
      try{
        Utilities.debugFSMExecution("Current state: "+currentState);
        Edge nextStateEdge=FSM[currentState].getNextState(token);
        String outputGenerated=nextStateEdge.output();
        int nextState=nextStateEdge.tail();
        outputPattern=outputPattern+outputGenerated;
        Utilities.debugFSMExecution(" Input: "+token+" Next state: "+nextState+" Output: "+outputGenerated);
        currentState=nextState;
      }catch (NoNextStateException e){
        Utilities.printException("wpAlgorithm", "runFSM", " Invalid token: "+token);
      }
    }
    Utilities.debugFSMExecution("FSM execution completed.");
    Utilities.debugFSMExecution("Output pattern:"+outputPattern);
  }
  /**
   * 
   Unmark all states of the FSM.
   
   */
  
  public static void unMarkAll(){
    for (int i=startState; i<=numberOfStates; i++){
      FSM[i].unMark();
    }
  }// End of unmarkAll()
  /**
   * 
   Return the index of the elemtn with the shortest string.
   */
  
  private static int shortestLength(String [] paths, int numberOfPaths){
    int currentLength=paths[0].length();
    if(numberOfPaths==0)return(0);
    int index=0;
    for (int i=0;i<numberOfPaths; i++)
    {
      if (paths[i].length()<currentLength)
      {
        index=i;
        currentLength=paths[i].length();
      }
    }
    return(index);
  }
  
  
  
  public static boolean existsInVector(String searchString, Vector searchVector){
    for(int i = 0; i < searchVector.size(); i++){
      if(searchVector.get(i).toString().equals(searchString))
      {
        return true;
      } 
    }
    return false;
  }// End of existsInVector()  
   
   
   public static void printAllTestCases(testingTree tree, pTableManager tablemanager){
    Vector pVector = new Vector();
    Vector wVector = new Vector();
    Vector testCases = new Vector();
    pVector = tree.getPValues();
    wVector = tablemanager.getW();
    pVector.add(new String(""));
    
    for(int i = 0; i < pVector.size(); i++)
    {
      for(int j = 0; j < wVector.size(); j++)
      {
        String pValue = (String) pVector.get(i);
        String wValue = (String) wVector.get(j);
        String testCase = pValue + wValue;
        if(!existsInVector(testCase, testCases))
        {
          testCases.add(testCase);
        }
      }
    }// Emnd of printAllTestCases()
    
    System.out.println("\nNumber of Test Cases: " + testCases.size());
    System.out.println("Test cases: " + testCases);
    
  }// End of printAllTestCases()

   public static void getW(){
    
    Set Wi = new HashSet();
    Set Wj = new HashSet();
    Set EdgesToRemove = new HashSet();
    Set W = new HashSet();
    
    for(int i = 1; i <= numberOfStates; i++){
      
      System.out.println("Computing Wi for State:" + i);
      Wi = new HashSet(FSM[i].getEdgeSet());  
      
      for(int j = 1; j <= numberOfStates; j++)
      {
        
        
        if(i != j)
        {
          
          Wj = new HashSet(FSM[j].getEdgeSet());
          Iterator jIterator = Wj.iterator();
          
          while(jIterator.hasNext())
          {
            
            Edge e = (Edge) jIterator.next();
            
            
            Iterator wiIter = Wi.iterator();
            
            while(wiIter.hasNext())
            {
              Edge wiEdge = (Edge) wiIter.next();
              //System.out.println("Wi Edge: " + wiEdge.input() + "/" + wiEdge.output());
              //System.out.println("j Edge: " + e.input() + "/" + e.output());
              
              if((wiEdge.input().equals(e.input())) && (wiEdge.output().equals(e.output())))
              {
                System.out.println("EDGES ARE EQUAL.  REMOVING " + wiEdge.input() + "/" + wiEdge.output() + " FROM Wi.");
                wiIter.remove();
              }
            }  
          }//while  
        }//if
      }//for

      Iterator wiIterator = Wi.iterator();
      while(wiIterator.hasNext()){
        Edge edge = (Edge) wiIterator.next();
        System.out.println("State: " + i + " Wi: " + edge.input() + "/" + edge.output());
      }//while
    }
  }// End of getW()
   
   public static void sortInputs(){
    boolean Swapped = true;
    int csize = countInputs;
    
    while(Swapped){
      Swapped = false;
      for(int i = 0; i < csize-1; i++){
        if(inputArray[i].compareTo(inputArray[i+1]) > 0){
           Utilities.debugSort("Swapping " + inputArray[i] + " and " + inputArray[i+1]);
          String temp;
          temp = inputArray[i];
          inputArray[i] = inputArray[i+1];
          inputArray[i+1] = temp;
          Swapped = true;
        }
      }
      
      csize--;
    }
    
    Utilities.debugSort("SORTED INPUT ARRAY");
    for(int i = 0; i < countInputs; i++){
       Utilities.debugSort("Position:" + i + " " + inputArray[i]);
    }
  }// End of sortInput()
  
   /* 
   Driver for the W-algorithm.
   */
  public static void main(String [] args){
    

    System.out.println("Test Generation Using the W-method. V2.0. July 30, 2013\n");
    
    fileSource=new Scanner(System.in);
    
    fsmFilename=getFilename(); // Get  from the user file name for FSM.
     
    for (int i = 0; i < outputArray.length; i++){
      outputArray[i] = ""; // Initialize array that will contain the output alphabet.
    }
    
    for (int j = 0; j < inputArray.length; j++){
      inputArray[j] = ""; // Initialize array that will contain the input alphabet.
    }
    
    getFSM(); // Get the FSM. The input and output alphabets are determined in getFSM().
    
    String [] realInput = new String [countInputs];
  
    //sortInputs();
    for(int z = 0; z < realInput.length; z++){
      realInput[z] = inputArray[z]; // Real Input contains only the elements of the input alphabet.
    }
    Arrays.sort(realInput);
    System.out.println("FSM input from:  "+fsmFilename);
    if(Utilities.fsmPrintSw)
      printFSM(realInput); // Print FSM.
    
    // Generate testingTree (Section 5.6.3 in the textbook).
    testingTree test = new testingTree(FSM, numberOfStates); 
    // Generate P-tables and the W set (Section 5.5 in the textbook).
    pTableManager test2 = new pTableManager(FSM, numberOfStates, realInput);
    printAllTestCases(test, test2);
  
  }// End of main()
  
}//End of class wpAlgorithm
 




